@@ -15,7 +15,7 @@ except: from pdb import set_trace as stop # python debugger
 # === START print ===
 # custom print() function override to show caller (module and function) information before actual print string,
 # plus options colourised output
-def print(*args, error=False, warn=False, highlight=False, color="", **kwargs):
+def print(*args, error=False, warn=False, highlight=False, color="", no_prefix=False, **kwargs):
     from builtins import print as builtin_print
     from inspect import currentframe, getouterframes
     from colorama import Fore, Style
@@ -29,7 +29,7 @@ def print(*args, error=False, warn=False, highlight=False, color="", **kwargs):
     ind = prefix.find("<module>: ") # remove '<module>: ' from prefix
     if ind < 0: ind = len(prefix)
     prefix = prefix[:ind]
-    builtin_print(prefix, end="")
+    if not no_prefix: builtin_print(prefix, end="")
     message = "" # default message (can also be 'ERROR: ' or 'WARNING: ')
     color_str = "" # default color of text
     if color == "black"  : color_str = f"{Fore.BLACK}"
@@ -67,11 +67,12 @@ def print(*args, error=False, warn=False, highlight=False, color="", **kwargs):
 # === END print ===
 
 # === START plot ===
-def plot(x=None, y=None, xerr=None, yerr=None, type=None, xlabel='x', ylabel='y', label=None, marker=None, linestyle=None, linewidth=None,
+def plot(y=None, x=None, yerr=None, xerr=None, type=None, xlabel='x', ylabel='y', label=None, marker=None, linestyle=None, linewidth=None,
          show=False, pause=None, xlog=False, ylog=False, xlim=None, ylim=None, save=None, legend_loc='upper left', text=None, shaded_err=None, *args, **kwargs):
     import matplotlib.pyplot as plt
     if linestyle == 'long dashed': linestyle = (0,(5,5))
-    if x is not None and y is not None:
+    if y is not None:
+        if x is None: x = np.arange(len(y)) # in this case, we just use the array indices as the x axis values
         if text is not None:
             plt.text(x, y, text, *args, **kwargs)
         else:
@@ -196,19 +197,6 @@ def fit(func, xdata, ydata, weights=None, params=None, verbose=1, *args, **kwarg
 
 # === END fit ===
 
-# Decorator to have a function return a quantity with given unit
-def use_unit(unit):
-    from pint import UnitRegistry
-    use_unit.ureg = UnitRegistry()
-    def decorator_use_unit(func):
-        from functools import wraps
-        @wraps(func)
-        def wrapper_use_unit(*args, **kwargs):
-            value = func(*args, **kwargs)
-            return value * use_unit.ureg(unit)
-        return wrapper_use_unit
-    return decorator_use_unit
-
 # === Decorator to print the runtime of the decorated function ===
 def timer_decorator(func):
     from functools import wraps
@@ -249,10 +237,10 @@ class timer:
         self.dt = None
         self.start()
     def start(self):
-        self.start_time = datetime.now()
+        self.start_time = self.datetime.now()
         if not self.quiet: print("timer('"+self.name+"'): start time = "+str(self.start_time))
     def stop(self):
-        self.stop_time = datetime.now()
+        self.stop_time = self.datetime.now()
         if not self.quiet: print("timer('"+self.name+"'): stop time = "+str(self.stop_time))
     def get_dt(self):
         # check whether stop() was called; if not, call it here
@@ -265,14 +253,14 @@ class timer:
               ", dt = "+str(self.dt), ", dt (seconds) = "+str(self.dt.total_seconds())+" s")
 
 # === executes a shell command: input string 'cmd'
-def run_shell_command(cmd, quiet=False, print_only=False, **kargs):
+def run_shell_command(cmd, quiet=False, print_only=False, capture=False, **kargs):
     from subprocess import run
     if (not quiet) or print_only:
         if 'color' not in kargs.keys():
             kargs['color'] = 'magenta' # set default colour for shell command print
         print(cmd, **kargs)
     if (not print_only):
-        sp_result = run(cmd, shell=True)
+        sp_result = run(cmd, capture_output=capture, shell=True)
         return sp_result
 
 def check_for_overwrite(filename):
@@ -432,45 +420,78 @@ def get_kde_sample(data, n=1000, seed=1, show=False):
         plt.show(block=True)
     return data_resampled
 
-# === return log10(x), but with the sign preserved ===
-def log10(x):
-    ret = np.copy(x)
-    ind = ret != 0
-    sign = np.sign(ret)
-    ret[ind] = sign[ind]*np.log10(np.abs(ret[ind]))
-    return ret
-
 # === return x rounded to nfigs significant figures ===
 def round(x, nfigs=3, str_ret=False):
-    ret = float(np.format_float_positional(x, precision=nfigs, unique=False, fractional=False, trim='k'))
-    if str_ret: ret = str(ret)
+    # prepare function for scalar or array input and copy x into ret
+    ret = np.array(x)
+    scalar_input = False
+    if ret.ndim == 0:
+        ret = ret[None] # makes x 1D
+        scalar_input = True
+    # iterate over each element in ret
+    with np.nditer(ret, op_flags=['readwrite']) as it:
+        for elem in it:
+            elem[...] = float(np.format_float_positional(elem, precision=nfigs, unique=False, fractional=False, trim='k'))
+    # convert to string if requested
+    if str_ret: ret = ret.astype(str)
+    # strip dimension if input was scalar and return
+    if scalar_input:
+        ret = np.squeeze(ret)
+        if str_ret: return str(ret)
     return ret
 
 # round a value and its error (uncertainty) to given nfigs significant figures
 def round_with_error(val, val_err, nfigs=2):
-    n = int(np.log10(val_err)) # displacement from ones place
-    if val_err >= 1: n += 1
-    scale = 10**(nfigs - n)
-    val = np.round(val * scale) / scale
-    val_err = np.round(val_err * scale) / scale
-    return val, val_err
+    # prepare function for scalar or array input and copy into ret
+    ret_val = np.array(val)
+    ret_val_err = np.array(val_err)
+    if ret_val.shape != ret_val_err.shape:
+        print("input value and associated error must have the same shape", error=True)
+    scalar_input = False
+    if ret_val.ndim == 0:
+        ret_val = ret_val[None] # makes x 1D
+        ret_val_err = ret_val_err[None] # makes x 1D
+        scalar_input = True
+    # iterate over each element in ret
+    for i, val_err in enumerate(np.nditer(ret_val_err)):
+        n = int(np.log10(val_err)) # displacement from ones place
+        if val_err >= 1: n += 1
+        scale = 10**(nfigs - n)
+        ret_val[i] = np.round(ret_val[i] * scale) / scale
+        ret_val_err[i] = np.round(val_err * scale) / scale
+    # strip dimension if input was scalar and return
+    if scalar_input: return np.squeeze(ret_val), np.squeeze(ret_val_err)
+    return ret_val, ret_val_err
 
 # === return x in E-format ===
 def eform(x, prec=10, print_leading_plus=False):
     import decimal
-    xx = decimal.Decimal(float(x))
-    tup = xx.as_tuple()
-    xx = xx.quantize(decimal.Decimal("1E{0}".format(len(tup[1])+tup[2]-prec-1)), decimal.ROUND_HALF_UP)
-    tup = xx.as_tuple()
-    exp = xx.adjusted()
-    sign = '-' if tup.sign else '+' if print_leading_plus else ''
-    dec = ''.join(str(i) for i in tup[1][1:prec+1])
-    if prec > 0:
-        return '{sign}{int}.{dec}E{exp:+03d}'.format(sign=sign, int=tup[1][0], dec=dec, exp=exp)
-    elif prec == 0:
-        return '{sign}{int}E{exp:+03d}'.format(sign=sign, int=tup[1][0], exp=exp)
-    else:
-        return None
+    def eform_scalar(x):
+        xx = decimal.Decimal(float(x))
+        tup = xx.as_tuple()
+        xx = xx.quantize(decimal.Decimal("1E{0}".format(len(tup[1])+tup[2]-prec-1)), decimal.ROUND_HALF_UP)
+        tup = xx.as_tuple()
+        exp = xx.adjusted()
+        sign = '-' if tup.sign else '+' if print_leading_plus else ''
+        dec = ''.join(str(i) for i in tup[1][1:prec+1])
+        if prec > 0:
+            return '{sign}{int}.{dec}E{exp:+03d}'.format(sign=sign, int=tup[1][0], dec=dec, exp=exp)
+        elif prec == 0:
+            return '{sign}{int}E{exp:+03d}'.format(sign=sign, int=tup[1][0], exp=exp)
+        else:
+            return None
+    # prepare function for scalar or array input and copy into ret
+    ret = np.array(x).astype(str)
+    scalar_input = False
+    if ret.ndim == 0:
+        ret = ret[None] # makes x 1D
+        scalar_input = True
+    # iterate over each element in ret
+    for i, val in enumerate(np.nditer(ret)):
+        ret[i] = eform_scalar(val)
+    # strip dimension if input was scalar and return
+    if scalar_input: return str(np.squeeze(ret))
+    return ret
 
 # escape latex
 def tex_escape(text):
@@ -493,7 +514,13 @@ def tex_escape(text):
     return regex.sub(lambda match: conv[match.group()], text)
 
 # === read an ASCII file ===
-def read_ascii(filename, astropy_read=True, read_header=True, quiet=False, max_num_lines=1e7, *args, **kwargs):
+# === select plain=True, to read each line into a list ===
+def read_ascii(filename, astropy_read=True, read_header=True, quiet=False, max_num_lines=1e7, plain=False, *args, **kwargs):
+    if plain:
+        with open(filename) as file:
+            lines = [line.rstrip() for line in file]
+        return lines
+    # ASCII table reading
     from astropy.io.ascii import read as astropy_ascii_read
     from astropy.table import Table
     if not quiet: print("reading data in '"+filename+"'...")
@@ -516,10 +543,17 @@ def read_ascii(filename, astropy_read=True, read_header=True, quiet=False, max_n
     return tab
 
 # === write an ASCII file ===
-def write_ascii(filename, dat, format='fixed_width', delimiter="", comment=False, quiet=False, *args, **kwargs):
-    from astropy.io.ascii import write as astropy_ascii_write
-    astropy_ascii_write(dat, filename, overwrite=True, format=format, delimiter=delimiter, comment=comment, *args, **kwargs)
-    if not quiet: print("File '"+filename+"' written; (nrow,ncol) = ({:d},{:d}).".format(len(dat), len(dat.columns)))
+# === select plain=True, to write list (dat) with lines to file ===
+def write_ascii(filename, dat, format='fixed_width', delimiter="", comment=False, quiet=False, plain=False, *args, **kwargs):
+    if plain:
+        with open(filename, 'w') as file:
+            for line in dat:
+                file.write(f"{line}\n")
+    else:
+        from astropy.io.ascii import write as astropy_ascii_write
+        astropy_ascii_write(dat, filename, overwrite=True, format=format, delimiter=delimiter, comment=comment, *args, **kwargs)
+        if not quiet: print("Table written with (nrow, ncol) = ({:d},{:d}).".format(len(dat), len(dat.columns)))
+    if not quiet: print("File '"+filename+"' written.")
 
 # smoothing/filtering data
 def smooth(x, y, window_npts=11, order=3):
@@ -548,8 +582,6 @@ def r_larmor(B, v, m=const.m_p, q=const.ec):
 def t_larmor(B, m=const.m_p, q=const.ec):
     return 2 * np.pi * m / (q * B)
 
-# === return freefall time ===
-@use_unit("s")
 def tff(rho):
     return np.sqrt(3.0*np.pi/(32.0*const.g_n*rho))
 
